import type { DenoResolverOptions } from "./denoResolver.ts";
import { denoResolver } from "./denoResolver.ts";

export type { DenoResolverOptions };
export { denoResolver };
export default denoResolver;
